/*
 * @(#)GlassfishInplaceMojo.java
 * Date 2012-07-20
 * Version 1.0
 * Author Jim Horner
 * Copyright (c)2012
 */


package us.hornerscorners.mojo.glassfish;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.InputStreamReader;

import java.nio.file.Files;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.io.FileUtils;

import org.apache.maven.artifact.Artifact;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;

import org.apache.maven.project.MavenProject;

import org.glassfish.embeddable.Deployer;
import org.glassfish.embeddable.GlassFish;
import org.glassfish.embeddable.GlassFishException;
import org.glassfish.embeddable.GlassFishProperties;
import org.glassfish.embeddable.GlassFishRuntime;

import org.glassfish.embeddable.web.ConfigException;
import org.glassfish.embeddable.web.VirtualServer;
import org.glassfish.embeddable.web.WebContainer;

import org.glassfish.embeddable.web.config.VirtualServerConfig;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * @goal run
 * @requiresDependencyResolution compile
 */
public class GlassfishInplaceMojo extends AbstractMojo {

    /**
     * @parameter
     */
    private File[] classpathItems;

    /**
     * @parameter
     */
    private File[] classpathOverlays;

    /**
     * @parameter
     */
    private String[] deployParameters;

    /**
     * @parameter default-value="${project.build.directory}/glassfish-inplace"
     */
    private File destination;

    /**
     * @parameter expression="${project}"
     */
    private MavenProject mavenProject;

    /**
     * @parameter default-value="8080"
     */
    private Integer port;

    /** Field description */
    private final FileFilter versionFilter = new VersionControlFilter();

    /** Field description */
    private final Logger logger = LoggerFactory.getLogger(getClass());

    /**
     * @parameter default-value="${basedir}/src/main/webapp"
     */
    private File webappDir;

    /**
     * @parameter
     */
    private File[] webinfOverlays;

    /**
     * Method description
     *
     *
     * @return
     */
    public List<File> buildLib() {

        List<File> result = new ArrayList<>();

        for (Object obj : this.mavenProject.getCompileArtifacts()) {

            Artifact artifact = (Artifact) obj;

            if (Artifact.SCOPE_COMPILE.equals(artifact.getScope())) {

                File file = artifact.getFile();

                // logger.error("Adding classpath item {}", file);
                result.add(file);
            }
        }

        return result;
    }

    /**
     * Method description
     *
     *
     */
    private void copyContents(File dir, File dest) throws IOException {

        File[] itemfiles = dir.listFiles(versionFilter);

        for (File item : itemfiles) {

            if (item.isDirectory()) {

                FileUtils.copyDirectoryToDirectory(item, dest);

            } else {

                FileUtils.copyFileToDirectory(item, dest);
            }
        }
    }

    /**
     * Method description
     *
     *
     * @param files
     * @param dest
     *
     * @throws IOException
     */
    private void copyFiles(File[] files, File dest) throws IOException {

        if ((files != null) && (files.length > 0)) {

            for (File file : files) {

                FileUtils.copyFileToDirectory(file, dest);
            }
        }
    }

    /**
     * Method description
     *
     *
     * @param fromWebInf
     * @param toWebInf
     */
    private void copyWebInfContents(File fromWebInf, File toWebInf)
            throws IOException {

        File[] itemfiles = fromWebInf.listFiles(versionFilter);

        for (File item : itemfiles) {

            if (item.isDirectory()) {

                // link subfolders
                String filename = item.getName();
                File lnkfile = new File(toWebInf, filename);

                if (filename.length() > 0) {

                    Files.createSymbolicLink(lnkfile.toPath(), item.toPath());
                }

            } else {

                FileUtils.copyFileToDirectory(item, toWebInf);
            }
        }
    }

    /**
     * Method description
     *
     *
     * @param glassfish
     *
     * @throws GlassFishException
     */
    private String deployDocRoot(GlassFish glassfish)
            throws GlassFishException, IOException {

        if (this.destination.exists()) {

            FileUtils.forceDelete(this.destination);
        }

        FileUtils.forceMkdir(this.destination);

        File webinffile = new File(this.destination, "WEB-INF");

        FileUtils.forceMkdir(webinffile);

        File[] webappfiles = this.webappDir.listFiles(versionFilter);

        for (File file : webappfiles) {

            String filename = file.getName();

            if ((filename.length() > 0)) {

                File lnkfile = new File(this.destination, filename);

                // symlink everything but WEB-INF, copy those contents
                if (filename.equals("WEB-INF")) {

                    copyWebInfContents(file, lnkfile);

                } else {

                    Files.createSymbolicLink(lnkfile.toPath(), file.toPath());
                }
            }
        }

        // copy classes
        File classesfile = new File(webinffile, "classes");

        FileUtils.forceMkdir(classesfile);

        if (this.classpathItems != null) {

            for (File file : this.classpathItems) {

                if (file.isDirectory()) {

                    copyContents(file, classesfile);

                } else {

                    FileUtils.copyFileToDirectory(file, classesfile);
                }
            }
        }

        // copy overlays
        copyFiles(this.classpathOverlays, classesfile);
        copyFiles(this.webinfOverlays, webinffile);

        // copy lib jars
        File libfile = new File(webinffile, "lib");

        FileUtils.forceMkdir(libfile);

        for (File file : buildLib()) {

            FileUtils.copyFileToDirectory(file, libfile);
        }

        // deploy new folder
        Deployer deployer = glassfish.getDeployer();

        return deployer.deploy(this.destination, this.deployParameters);
    }

    /**
     * Method description
     *
     *
     * @throws MojoExecutionException
     */
    @Override
    public void execute() throws MojoExecutionException {

        try {

            GlassFishProperties gfProps = new GlassFishProperties();

            gfProps.setPort("http-listener", this.port);

            GlassFish glassfish =
                GlassFishRuntime.bootstrap().newGlassFish(gfProps);

            ShutdownHook shutdownHook = new ShutdownHook(glassfish,
                                            this.destination.getName());

            Runtime.getRuntime().addShutdownHook(shutdownHook);

            glassfish.start();

            WebContainer webContainer =
                glassfish.getService(WebContainer.class);

            VirtualServer virtualServer =
                webContainer.getVirtualServer("server");

            VirtualServerConfig vscfg = virtualServer.getConfig();

            if (vscfg == null) {
                vscfg = new VirtualServerConfig();
            }

            vscfg.setAllowLinking(true);
            virtualServer.setConfig(vscfg);

            deployDocRoot(glassfish);

            waitForKey();

            shutdownHook.executeHook();

        } catch (GlassFishException | IOException | ConfigException e) {

            throw new MojoExecutionException("Error", e);
        }

    }

    /**
     * Method description
     *
     *
     * @throws IOException
     */
    private void waitForKey() throws IOException {

        String key = "";

        System.out.println("Press X to exit");

        BufferedReader br = null;

        try {

            br = new BufferedReader(new InputStreamReader(System.in));

            while ("X".equals(key) == false) {

                key = br.readLine();
            }

        } finally {

            if (br != null) {

                try {

                    br.close();

                } catch (IOException e) {

                    // nothing can be done
                }
            }
        }
    }
}
